

import java.io.*;
import java.util.*;
 
/** Class for storing and accessing .rel system relation in memory.
 *
 * The .rel system relation is as follows:
 * <pre>
 * .rel(.name   .numattribs   .numtuples   .vrc   .tree   .persistent)
 *
 * .name          identifier for the relation
 * .numattribs    number of attributes the relation is defined on
 * .numtuples     number of tuples in the relation
 * .vrc           indicates whether this is a view, a relation or a comp
 * .tree          root of syntax tree of code if virtual, NULL otherwise
 * .persistent    boolean flag indicating if this relation is persistent
 * </pre>
 * Some information from the .rd system table is kept in class RelTable
 * as well. This is the information describing what domains belong to
 * each relation.
 * The .rd system relation is as follows:
 * <pre>
 * .rd(.relname  .domname  .count  .dompos .sortrank)
 *
 * .name          identifier for the relation or domain
 * .attribute     identifier for an attribute of the relation or domain
 * .drc           indicates if this identifier is a domain, relation or a comp
 * .count         number of different values of the domain in this relation
 * .dompos        index of first byte of domain in the relation
 * .sortrank      position of domain in sort list
 * </pre>
 *
 * @see     Relation
 */ 
//public class RelTable implements TypeConstants //modified by Zongyan
public class RelTable implements TypeConstants,Constants
{
   /**
	* Constants declaration.
	*/
   private static final int    BUF_SIZE  = 16 * 1024;
   private static final String REL_NAME  = ".rel";
   private static final String RD_NAME   = ".rd";
   private static final String VIEW_NAME = ".view";
   private static final String COMP_NAME = ".comp";

   /**
	* A hash table storing all the relation entries.
	*/
   private Hashtable table;


   /**
	* Initialize the hash table, and load relation information
	* from disk files .rel & .rd if requested. Loading is only
	* needed at system startup for the top-level global frame.
	*/
   public RelTable(Environment env, boolean load)
   {  table = new Hashtable();
	  if(load)
	  {  this.load(env);
	  }
   }      
   /**
	* Add new Relations without putting in the actual domain entries. This
	* method is supposed to be called by load() method by Interpreter only.
	* @param   relnames   The names of the new domains
	* @param   numTuples The number of tuples in this relation
	* @param   numAttr  The number of attributes in this relation
	* @param   rvc    The type of the new relation
	* @param   sort  The number of sort attributes 
	*/
   private void addRel(String name, int numTuples, int numAttr,
					   int rvc, int sort, Environment env)
   {   Domain doms[];

	  // If this relation is defined already, remove it first. 
	  if(isRel(name)) rmRel(name);

	  // We only put in a skeleton domain array here. The insertIDlist()
	  // method in this class will insert the domain entries.
	  doms = new Domain[numAttr];

//modified by Yi Zheng March 22, 2002 to handle startup .rel, .com not found
	  table.put(name, new Relation(name, doms, rvc, numTuples,
								   numAttr, sort, (SimpleNode) null, env));
   }      
   public void addRel(Relation rel, boolean incNumRef)
   {  String name = rel.name;
	  // If this relation is defined already, remove it first. 
	  if(isRel(name))
	  {  rmRel(name);
	  }

	  if(incNumRef)
	  {  int attributes; 
		 if (rel.domains == null) attributes = 0;
		 else attributes = rel.domains.length;

		 for(int i=0; i<attributes; i++)
		 {  rel.domains[i].numref++;
		 }
	  }
	  table.put(name, rel);
   }      
   /**
	* Dump RelTable to the files .rel, .rd and .comp
	*/
/*
   public void dump()
   {  try
	  {
		 BlockOutputStream relos = new BlockOutputStream(REL_NAME,BUF_SIZE);
		 BlockOutputStream rdos = new BlockOutputStream(RD_NAME,BUF_SIZE);
		 BlockOutputStream compos = new BlockOutputStream(COMP_NAME,BUF_SIZE);

		 for(Enumeration e = table.keys(); e.hasMoreElements();)
		 {  Relation relent = (Relation)table.get(e.nextElement());
			//No need to dump any info about computations
			//All the info that is needed will be stored in .expr
			//See load() method of class ExprTable to see how 
			//computations are loaded when the system boots.
			if(relent.rvc == RELATION || relent.rvc == VIEW)
			{  // Dump relation info. to .rel.
			   relos.writeString(relent.name);
			   relos.writeInt(relent.numtuples);
			   relos.writeInt(relent.numattrs);
			   relos.writeInt(relent.rvc);
			   relos.writeInt(relent.numsortattrs);

			   // Dump the rel-dom info. to .rd file.
			   for(int i = 0; i < relent.numattrs; i++)
			   {  rdos.writeString(relent.name);
				  rdos.writeString(relent.domains[i].name);
				  rdos.writeInt(i);
			   }
			   // Note, we don't need to take care of view here, since
			   // the exprTable will do this job.
			}
		 }

		 // close the output file streams.
		 relos.close();
		 rdos.close();
		 compos.close();
	  }
	  catch(IOException e)
	  {  // Shouldn't happen..
		 Global.println("error: RelTable: dump IOException");

		 // There must an error we didn't catch. If control comes here,
		 // we must figure out what was going on and fix it.
		 System.exit(1);
	  }
   }         
   */


   /**
	* Dump RelTable to the files .rel, .rd and .comp 
	* 	-- modified by Weizhong Sun, Jun 14th 1999
	*/
   public void dump()
   {  try
	  {
		 BlockOutputStream relos = new BlockOutputStream(REL_NAME, BUF_SIZE);
		 BlockOutputStream rdos = new BlockOutputStream(RD_NAME, BUF_SIZE);
		 BlockOutputStream compos = new BlockOutputStream(COMP_NAME, BUF_SIZE);//not used here Yi Zheng March 22, 2002
		 String wb = new String("");

		 for(Enumeration e = table.keys(); e.hasMoreElements();)
		 {  Relation relent = (Relation)table.get(e.nextElement());
			//No need to dump any info about computations
			//All the info that is needed will be stored in .expr
			//See load() method of class ExprTable to see how 
			//computations are loaded when the system boots.
			if(relent.rvc == RELATION || relent.rvc == VIEW)
			{ 
			   // Dump relation info. to .rel.
			   relos.writeMyBytes(STRING, relent.name);
//			   relos.writeBytes("^F");	// delimiter
			   relos.writeByte(6);	// delimiter
			   wb = wb.valueOf(relent.numtuples);
			   relos.writeMyBytes(INTEGER, wb);
//			   relos.writeBytes("^F");
			   relos.writeByte(6);	// delimiter
			   wb = wb.valueOf(relent.numattrs);
			   relos.writeMyBytes(INTEGER, wb);
//			   relos.writeBytes("^F");
			   relos.writeByte(6);	// delimiter
			   wb = wb.valueOf(relent.rvc);
			   relos.writeMyBytes(STRING, wb);
//			   relos.writeBytes("^F");
			   relos.writeByte(6);	// delimiter
			   wb = wb.valueOf(relent.numsortattrs);
			   relos.writeMyBytes(INTEGER, wb);
//			   relos.writeBytes("^F");
			   relos.writeByte(6);	// delimiter
//System.out.println("Dump active!!"+ relent.active);
//                           wb= wb.valueOf(relent.active);
//                           relos.writeMyBytes(INTEGER, wb);
//                           relos.writeByte(6);  // delimiter
			   relos.writeBytes("\n");	// new line

			   // Dump the rel-dom info. to .rd file.
			   for(int i = 0; i < relent.numattrs; i++)
			   { 
			  	rdos.writeMyBytes(STRING, relent.name);
//			   	rdos.writeBytes("^F");
			   	rdos.writeByte(6);
				rdos.writeMyBytes(STRING, relent.domains[i].name);
//			   	rdos.writeBytes("^F");
			   	rdos.writeByte(6);
				wb = wb.valueOf(i);
			  	rdos.writeMyBytes(INTEGER, wb);
//			   	rdos.writeBytes("^F");
			   	rdos.writeByte(6);
			   	rdos.writeBytes("\n");	
			   }
			   // Note, we don't need to take care of view here, since
			   // the exprTable will do this job.
			}
		 }

		 // close the output file streams.
		 relos.close();
		 rdos.close();
		 compos.close();
	  }
	  catch(IOException e)
	  {  // Shouldn't happen..
		 Global.println("error: RelTable: dump IOException");

		 // There must an error we didn't catch. If control comes here,
		 // we must figure out what was going on and fix it.
		 System.exit(1);
	  }
   }         

   /**
	* Fill the exprtable (a hashtable) before dumping the expressions
	* to .expr disk file.
	*/
   public void fillExprTable(Hashtable exprTable)
   {  String name;

	  for(Enumeration e = table.keys(); e.hasMoreElements();)
	  {  Relation relent = (Relation) table.get(e.nextElement());
		 if(relent.tree != null)
		 { 
//System.out.println("In RelTable, relent.tree.active = " + relent.tree.active);
			 exprTable.put(relent.name, new ExprEntry(relent.name,
						  (Object) relent.tree));
		 }
	  }
   }      
   /**
	*  Get the reference to Relation object of the relation 'name'.
	*  @param name The name of the relation.
	*/
	public Relation getRel(String name){
	   return((Relation) table.get(name));
   }      
   /**
	* Insert the idlist information from file .rd to RelTable.
	* This method is called only by Interpreter/Global (at start up).
	* @param relname The name of the relation
	* @param doms The array of the domain entries
	*/
   public void insertIDlist(String relname, Domain[] doms){

	  Relation relent = (Relation) table.get(relname);

	  if(relent != null){
	relent.domains = doms;

	// Note: No need to increment the reference counter
	// of this domain since the default is included the counter.
	  }else{
	// REVISIT: Shall we keep silent? seems no...
	// Maybe better to throw an exception..
	Global.println("error: RelTable.insertIDlist(): can't find "+
	   "the relation '"+relname+"' in reltable.");
	  }
   }         
   /**
	* Insert the idlist information from file .rd to RelTable.
	* This method is called only by Interpreter.
	* @param relname The name of the relation
	* @param domname The name of the domain
	* @param pos  The position of the domain in the relation
	*/
   public void insertIDlist(String relname, Domain doment, int pos)
   {  // REVISIT: pos is not taken care of currently. See loadRDTable()
	  // in Interpreter.java (Global.java)

	  Relation relent = (Relation) table.get(relname);

	  if(relent != null)
	  {  //relent must be for a relation or for a view
		 //since computations are not in the reltable yet
		 relent.domains[pos] = doment;

		 // Note: No need to increment the reference counter
		 // of this domain since the default is included the counter.
	  }
	  else
	  {  // REVISIT: Shall we keep silent?  seems no...
		 // Maybe better to throw an exception..
		 Global.println(
				  "Warning: RelTable.insertIDlist(): can't find "+
				  "the relation '"+relname+"' in reltable."  );
	  }
   }         
   /**
	* Check if a relation exists.
	* @param   name    The name of the relation to lookup
	* @return          True if name is already a relation, else false
	*/
   private boolean isRel(String name){
	  Relation r = (Relation) table.get(name);
	  return (r != null);
   }      
   public Enumeration keys()
   {  return table.keys();
   }      


   /**
	* Load the system tables .rel, and .comp into memory
	*/
/*
   public void load(Environment env)
   {  String name;
	  int numTuples;
	  int numAttr;
	  int rvc;
	  int numSort;

	  // load relation information from .rel.
	  try
	  {  BlockInputStream relis = new BlockInputStream(REL_NAME, BUF_SIZE);

		 // load data to hash table.
		 while(true)
		 {  // REVISIT: will need to handle bad record exceptions..
			try
			{  name = relis.readString();
			   numTuples = relis.readInt();
			   numAttr = relis.readInt();
			   rvc = relis.readInt();
			   numSort = relis.readInt();

			   // Add the relation or view without the domain entries.
			   // These will be inserted by loadRDTable() method in
			   // Interpreter.class
			   if (rvc == RELATION || rvc == VIEW)
			   {  addRel(name, numTuples, numAttr, rvc, numSort, env);
			   }
			   //NOTE: Computations will be put into the reltable 
			   //when the required information from the file .expr
			   //processed. See the load() method of class ExprTable.
			}
			catch(IOException e)
			{  relis.close();   // EOF
			   break;
			}
		 }
	  }
	  catch(FileNotFoundException e)
	  {  // REVISIT: not only here, but whenever the .rel is found to be
		 // corrupted.

		 // first time initialization....
		 // Add the relation without the domain entries. The 
		 // later will be inserted by loadRDTable() method in
		 // Interpreter.class
		 addRel(".rel", 3, 5, RELATION, 0, env);
		 addRel(".dom", 8, 3, RELATION, 0, env);
		 addRel(".rd", 10, 3, RELATION, 0, env);
	  }
	  catch(IOException e)
	  {  Global.println("error: RelTable: load() IOException");
	  }

	  // Note, we don't need to take care of loading view info. from
	  // .view here, since ExprTable will do its job. We also dont need
	  // to load .rd information here, since the interpreter does this 
	  // work.
   }         
   */

   /**
	* Load the ascii system files .rel, and .comp into memory
	* 	-- modified by Weizhong Sun, Jun 14th 1999
	*/
   public void load(Environment env)
   {  
   	  String name;
          int active;
	  int numTuples;
	  int numAttr;
	  int rvc;
	  int numSort;
	  Relation rel = new Relation();
	  String rb;
	  Integer tint;

	  // load relation information from .rel.
	  try
	  {  
	  	BlockInputStream relis = new BlockInputStream(REL_NAME, BUF_SIZE);

		 // load data to hash table.
		 while(true)
		 {  // REVISIT: will need to handle bad record exceptions..
			try
			{  
			   rb = rel.strRead(relis, STRING);
			   name = rb;
			   rb = rel.strRead(relis, INTEGER);
			   tint = new Integer(rb);
			   numTuples = tint.intValue();
			   rb = rel.strRead(relis, INTEGER);
			   tint = new Integer(rb);
			   numAttr = tint.intValue();
			   rb = rel.strRead(relis, INTEGER);
			   tint = new Integer(rb);
			   rvc = tint.intValue();
			   rb = rel.strRead(relis, INTEGER);
			   tint = new Integer(rb);
			   numSort = tint.intValue();
//                           rb = rel.strRead(relis, INTEGER);
//                           tint = new Integer(rb);
//                           active = tint.intValue();

			   // Add the relation or view without the domain entries.
			   // These will be inserted by loadRDTable() method in
			   // Interpreter.class
			   if (rvc == RELATION || rvc == VIEW)
			   {  addRel(name, numTuples, numAttr, rvc, numSort, env);
			   }
			   //NOTE: Computations will be put into the reltable 
			   //when the required information from the file .expr
			   //processed. See the load() method of class ExprTable.
			}
			catch(IOException e)
			{  relis.close();   // EOF
			   break;
			}
		 }
	  }
	  catch(FileNotFoundException e)
	  {  // REVISIT: not only here, but whenever the .rel is found to be
		 // corrupted.

		 // first time initialization....
		 // Add the relation without the domain entries. The 
		 // later will be inserted by loadRDTable() method in
		 // Interpreter.class 
		 // SHOULD BE Global class in the above line -- Yi Zheng
		 addRel(".rel", 3, 5, RELATION, 0, env);
	//	 addRel(".dom", 20, 3, RELATION, 0, env);//changed from 8 tuples
	
		//Yi Zheng April 29, 2002
		 addRel(".dom", 21, 4, RELATION, 0, env);//changed from 8 tuples

		 addRel(".rd", 12, 3, RELATION, 0, env); //changed from 10 tuples
	  }
	  catch(IOException e)
	  {  Global.println("error: RelTable: load() IOException");
	  }

	  // Note, we don't need to take care of loading view info. from
	  // .view here, since ExprTable will do its job. We also dont need
	  // to load .rd information here, since the interpreter does this 
	  // work.
   }         




   /**
	* Delete a relation from the RelTable.
	* @param   name   The name of the relation.
	*/
   public void rmRel(String name)
   {  // look for the relation we want to remove.
	  Relation r = (Relation) table.get(name);

	  // if the relation exists...
	  // REVISIT: this must be more complicated because of the 
	  // nested relation case..
	  
	  //need to handle cascading delete in case of a nested relation - Yi Zheng
	  if( r != null)
	  {  // decrement the reference counter of referenced domains
//System.out.println("Got the relation in table.");	
 		 int attributes;
		 if (r.domains == null) attributes = 0;
		 else attributes = r.domains.length;
		 for(int i = 0; i < attributes; i++)
		 {
		    r.domains[i].numref--;
		 }

		 // remove this relation from hashtable.
		 table.remove(name);

		 if(r.rvc == COMPUTATION)
		 {  Global.comptable.rmComps(name); //this does nothing Yi Zheng
                    // remove the data file of computation 
                    String fcompname;
                    fcompname = name + ".comp";
                    File fcomp = new File(fcompname);
                   // if (fcomp.exists()) fcomp.delete();
                   					//modified by Zongyan  -- Sep 2002
					if (fcomp.exists()) {
						if (!fcompname.startsWith("_temp")){
							 String mode = aldatpUtility.getmode(fcompname);
							 if (Global.DataFileSet.containsKey(fcompname))
								 Global.DataFileSet.remove(fcompname);
						     Global.DataFileSet.put(fcompname,mode);
						}
						fcomp.delete();
					}

		 }

		 // remove the data file of relation
		 File f = new File(name);
	//	 if (f.exists()) f.delete();
			 //moidified by Zongyan -- Sep 2002
		 if (f.exists()) {
			 if (!name.startsWith("_temp")){
				 String mode = aldatpUtility.getmode(name);
				 if (Global.DataFileSet.containsKey(name))
				 Global.DataFileSet.remove(name);
				 Global.DataFileSet.put(name,mode);
			}
			 f.delete();
		 }


	  }
	  else
	  {  Global.println("warning: relation '"+name+"' is not found in "
						 + "the relTable.");
	  }
   }         
   /**
	* Show a relation entry.
	*/
   public void showRel(Relation relent) {
	   PrettyPrint out = new PrettyPrint();

	  // print the title.
	  out.pprintRule(30);    out.pprint(" Relation Entry ", 16);
	  out.pprintRule(30);    out.println();
	  out.pprint("Name        Type          Arity       NTuples     Sort    Active", 65);
	  out.println();
	  out.pprintRule(86);
	  out.println();

	  out.pprint(relent.name, 12);
	  out.pprint(Utility.type2String(relent.rvc));
	  out.pprint(relent.numattrs);
	  out.pprint(relent.numtuples);
	  out.pprint(relent.numsortattrs);
          out.pprint(relent.active);

	  if(relent.tree != null){
		 out.println();
	//	 relent.tree.dump("                              "); modified by Yi Zheng March 23, 2002
		Utility.dump(relent.tree, "                              ");
	  }
	  out.println();

	  out.pprintRule(86);
	  out.println();
   }         
   /**
	* Show table.
	*/
   public void showTable(boolean flag) {
	   Relation relent;
	   PrettyPrint out = new PrettyPrint();

	  // print the title.
	  out.pprintRule(35);    out.pprint(" Relation Table ", 16);
	  out.pprintRule(35);    out.println();
	  out.pprint("Name                            Type          Arity       NTuples     Sort    Active", 85);
	  out.println();
	  out.pprintRule(86);
	  out.println();

	  for(Enumeration e = table.keys(); e.hasMoreElements(); ){
		 relent = (Relation) table.get(e.nextElement());

	if(flag || !relent.name.startsWith(".")){
	   out.pprint(relent.name, 30); // Weizhong, 12->30
	   out.pprint(Utility.type2String(relent.rvc));
	   out.pprint(relent.numattrs);
	   out.pprint(relent.numtuples);
	   out.pprint(relent.numsortattrs);
           out.pprint(relent.active);
/* 
//REVISIT: Patrick took this out because
//the output is too large for computations
		if(relent.tree != null){
		  out.println();
		  relent.tree.dump("                              ");
		  out.pprint("--");
			}
*/
	   out.println();
		 }
	  }
	  out.pprintRule(86);
	  out.println();
	  //out.close();
   }         

  public Hashtable getRelTable() {
	return table;
  }
  
  /**
   * function to provide number of tuples in .rel
   * note that we don't count computations, as they are not dumped into .rel
   * @author Yi Zheng March 22, 2002
   */
   public int getRelTableSize()
   {
   	int tSize = 0;
   	for ( Enumeration enum = table.keys();
    			enum.hasMoreElements(); )
    	{
    		Relation relent = (Relation)table.get(enum.nextElement());
    		if(relent.rvc == RELATION || relent.rvc == VIEW)
    			tSize++;
    	}
    return tSize;

   }
   
   /**
    * function to provide rdTable size
    * @author Yi Zheng March 22, 2002
    */
    public int getRDTableSize()
    {
    	int totSize = 0;
    	for ( Enumeration enum = table.keys();
    			enum.hasMoreElements(); )
    	{
    		Relation relent = (Relation)table.get(enum.nextElement());
    		if(relent.rvc == RELATION || relent.rvc == VIEW)
    			totSize += relent.numattrs;
    	}
    	return totSize;
    }
}
